@@ -7,13 +7,19 @@ needing an email (no password). The vehicle option codes are loaded from
 
 # Author: Tim Dorssers
 
+__version__ = '1.5.0'
+
 import os
+import re
+import ast
 import json
 import time
 import base64
 import hashlib
 import logging
 import pkgutil
+import tempfile
+import webbrowser
 try:
     from HTMLParser import HTMLParser
     from urlparse import urljoin
@@ -23,8 +29,9 @@ except ImportError:
 import requests
 from requests_oauthlib import OAuth2Session
 from requests.exceptions import *
+from requests.packages.urllib3.util.retry import Retry
 from oauthlib.oauth2.rfc6749.errors import *
-
+import websocket  # websocket-client v0.49.0 up to v0.58.0 is not supported
 
 requests.packages.urllib3.disable_warnings()
 
@@ -32,6 +39,7 @@ BASE_URL = 'https://owner-api.teslamotors.com/'
 CLIENT_ID = 'e4a9949fcfa04068f59abb5a658f2bac0a3428e4652315490b659d5ab3f35a9e'
 SSO_BASE_URL = 'https://auth.tesla.com/'
 SSO_CLIENT_ID = 'ownerapi'
+STREAMING_BASE_URL = 'wss://streaming.vn.teslamotors.com/'
 
 
 class PasswdFilter(logging.Filter):
@@ -62,28 +70,47 @@ logger.addHandler(logging.NullHandler())
 logger.addFilter(PasswdFilter())
 logging.getLogger('requests_oauthlib.oauth2_session').addFilter(PasswdFilter())
 
+# Py2/3 compatibility
+try:
+    input = raw_input
+except NameError:
+    pass
+
 
 class Tesla(requests.Session):
     """ Implements a session manager for the Tesla Motors Owner API
 
-    :param email: SSO identity.
-    :param password: SSO credential.
-    :passcode_getter: Function that returns the TOTP passcode.
-    :factor_selector: Function with one argument, a list of factor dicts, that
-                      returns the selected dict or factor name.
-    :param verify: Verify SSL certificate.
-    :param proxy: URL of proxy server.
+    email: SSO identity.
+    password: SSO credential.
+    passcode_getter: (optional) Function that returns the TOTP passcode.
+    factor_selector: (optional) Function with one argument, a list of factor
+        dicts, that returns the selected dict or factor name.
+    captcha_solver: (optional) Function with one argument, SVG image content,
+        that returns the captcha characters.
+    verify: (optional) Verify SSL certificate.
+    proxy: (optional) URL of proxy server.
+    retry: (optional) Number of connection retries or `Retry` instance.
+    user_agent: (optional) The User-Agent string.
+    cache_file: (optional) Path to cache file used by default loader and dumper.
+    cache_loader: (optional) Function that returns the cache dict.
+    cache_dumper: (optional) Function with one argument, the cache dict.
     """
 
     def __init__(self, email, password, passcode_getter=None,
-                 factor_selector=None, verify=True, proxy=None):
+                 factor_selector=None, captcha_solver=None, verify=True,
+                 proxy=None, retry=0, user_agent=__name__ + '/' + __version__,
+                 cache_file='cache.json', cache_loader=None, cache_dumper=None):
         super(Tesla, self).__init__()
         if not email:
             raise ValueError('`email` is not set')
         self.email = email
         self.password = password
-        self.passcode_getter = passcode_getter
-        self.factor_selector = factor_selector
+        self.passcode_getter = passcode_getter or self._get_passcode
+        self.factor_selector = factor_selector or self._select_factor
+        self.captcha_solver = captcha_solver or self._solve_captcha
+        self.cache_loader = cache_loader or self._cache_load
+        self.cache_dumper = cache_dumper or self._cache_dump
+        self.cache_file = cache_file
         self.token = {}
         self.expires_at = 0
         self.authorized = False
@@ -91,18 +118,20 @@ class Tesla(requests.Session):
         self.sso_token = {}
         self.sso_base = SSO_BASE_URL
         # Set Session properties
-        self.headers.update({'Content-Type': 'application/json'})
+        self.mount('https://', requests.adapters.HTTPAdapter(max_retries=retry))
+        self.headers.update({'Content-Type': 'application/json',
+                             'User-Agent': user_agent})
         self.verify = verify
         if proxy:
             self.trust_env = False
             self.proxies.update({'https': proxy})
         self._token_updater()  # Try to read token from cache
 
-    def request(self, method, uri, data=None, **kwargs):
+    def request(self, method, url, **kwargs):
         """ Extends base class method to support bearer token insertion. Raises
         HTTPError when an error occurs.
 
-        :rtype: JsonDict
+        Return type: JsonDict
         """
         # Auto refresh token and insert access token into headers
         if self.authorized:
@@ -110,10 +139,13 @@ class Tesla(requests.Session):
                 self.refresh_token()
             self.headers.update({'Authorization':
                                  'Bearer ' + self.token['access_token']})
-        # Construct URL, serialize data and send request
-        url = BASE_URL + uri.strip('/')
+        # Construct URL and send request with optional serialized data
+        url = BASE_URL + url.strip('/')
+        kwargs.setdefault('timeout', 10)
+        data = kwargs.pop('data', {})
         logger.debug('Requesting url %s using method %s', url, method)
         logger.debug('Supplying headers %s and data %s', self.headers, data)
+        logger.debug('Passing through key word arguments %s', kwargs)
         response = super(Tesla, self).request(method, url, json=data, **kwargs)
         # Error message handling
         if 400 <= response.status_code < 600:
@@ -144,6 +176,7 @@ class Tesla(requests.Session):
         oauth.verify = self.verify
         oauth.trust_env = self.trust_env
         oauth.proxies = self.proxies
+        oauth.headers['User-Agent'] = self.headers['User-Agent']
         url, _ = oauth.authorization_url(self.sso_base + 'oauth2/v3/authorize',
                                          code_challenge=code_challenge,
                                          code_challenge_method='S256',
@@ -154,22 +187,48 @@ class Tesla(requests.Session):
         if response.history:
             self.sso_base = urljoin(response.url, '/')
         # Parse input objects on HTML form
-        form = HTMLForm()
-        form.feed(response.text)
-        transaction_id = form.data['transaction_id']
+        form = HTMLForm(response.text)
+        transaction_id = form['transaction_id']
+        # Retrieve captcha image if required
+        if 'captcha' in form:
+            response = oauth.get(self.sso_base + 'captcha')
+            response.raise_for_status()  # Raise HTTPError, if one occurred
+            form['captcha'] = self.captcha_solver(response.content)
+            if not form['captcha']:
+                raise ValueError('Missing captcha response')
         # Submit login credentials to get authorization code through redirect
-        form.data.update({'identity': self.email, 'credential': self.password})
+        form.update({'identity': self.email, 'credential': self.password})
         response = oauth.post(self.sso_base + 'oauth2/v3/authorize',
-                              data=form.data, allow_redirects=False)
-        response.raise_for_status()  # Raise HTTPError, if credentials invalid
-        if response.status_code == 200:
-            # Check if login form is on page, cause for example locked account
-            form = HTMLForm()
-            form.feed(response.text)
-            if form.data:
-                raise ValueError('Credentials rejected')
+                              data=form, allow_redirects=False)
+        if response.status_code != 302:
+            form = HTMLForm(response.text)
+            # Retrieve captcha image if required
+            if 'captcha' in form:
+                response = oauth.get(self.sso_base + 'captcha')
+                response.raise_for_status()  # Raise HTTPError, if one occurred
+                form['captcha'] = self.captcha_solver(response.content)
+                if not form['captcha']:
+                    raise ValueError('Missing captcha response')
+                # Resubmit login credentials
+                form.update({'identity': self.email, 'credential': self.password})
+                response = oauth.post(self.sso_base + 'oauth2/v3/authorize',
+                                      data=form, allow_redirects=False)
+        if response.status_code != 302:
+            # An error occurred if the login form is present
+            msgs = ['Credentials rejected'] if HTMLForm(response.text) else []
+            # Look for messages object literal
+            m = re.search(r'var messages = ({.*});', response.text)
+            if m:
+                # Make list of error messages
+                for msg in json.loads(m.group(1)).values():
+                    msgs += [msg] if not isinstance(msg, list) else msg
+            # Raise error if one occurred
+            if msgs:
+                raise ValueError('. '.join(msgs))
+            response.raise_for_status()
             # Check for MFA factors to handle to get authorized
             response = self._check_mfa(oauth, transaction_id)
+        response.raise_for_status()
         # Use authorization response code in redirected location to get token
         url = response.headers.get('Location')
         oauth.fetch_token(self.sso_base + 'oauth2/v3/token',
@@ -187,13 +246,9 @@ class Tesla(requests.Session):
         factors = response.json()['data']
         if not factors:
             raise ValueError('No registered factors')
-        if not self.passcode_getter:
-            raise ValueError('`passcode_getter` callback is not set')
         if len(factors) == 1:
             factor = factors[0]  # Auto select only factor
         elif len(factors) > 1:
-            if not self.factor_selector:
-                raise ValueError('`factor_selector` callback is not set')
             # Get selected factor
             factor = self.factor_selector(factors)
             if not factor:
@@ -245,25 +300,57 @@ class Tesla(requests.Session):
         self.authorized = True
         self._token_updater()  # Save new token
 
-    def _token_updater(self):
-        """ Handles token persistency """
-        # Open cache file
+    @staticmethod
+    def _get_passcode():
+        """ Default passcode_getter method """
+        return input('Passcode: ')
+
+    @staticmethod
+    def _select_factor(factors):
+        """ Default factor_selector method """
+        for i, factor in enumerate(factors):
+            print('{:2} {}'.format(i, factor['name']))
+        return factors[int(input('Select factor: '))]
+
+    @staticmethod
+    def _solve_captcha(svg):
+        """ Default captcha_solver method """
+        # Use web browser to display SVG image
+        with tempfile.NamedTemporaryFile(suffix='.svg', delete=False) as f:
+            f.write(svg)
+        logger.debug('Opening %s with default browser', f.name)
+        webbrowser.open('file://' + f.name)
+        return input('Captcha: ')
+
+    def _cache_load(self):
+        """ Default cache loader method """
         try:
-            with open('cache.json') as infile:
+            with open(self.cache_file) as infile:
                 cache = json.load(infile)
         except (IOError, ValueError):
             cache = {}
-        # Write token to cache file
+        return cache
+
+    def _cache_dump(self, cache):
+        """ Default cache dumper method """
+        try:
+            with open(self.cache_file, 'w') as outfile:
+                json.dump(cache, outfile)
+        except IOError:
+            logger.error('Cache not updated')
+        else:
+            logger.debug('Updated cache')
+
+    def _token_updater(self):
+        """ Handles token persistency """
+        cache = self.cache_loader()
+        if not isinstance(cache, dict):
+            raise ValueError('`cache_loader` must return dict')
+        # Write token to cache
         if self.authorized:
             cache[self.email] = {'url': self.sso_base, 'sso': self.sso_token,
                                  SSO_CLIENT_ID: self.token}
-            try:
-                with open('cache.json', 'w') as outfile:
-                    json.dump(cache, outfile)
-            except IOError:
-                logger.error('Cache not updated')
-            else:
-                logger.debug('Updated cache')
+            self.cache_dumper(cache)
         # Read token from cache
         elif self.email in cache:
             self.sso_base = cache[self.email].get('url', SSO_BASE_URL)
@@ -302,7 +389,7 @@ class Tesla(requests.Session):
         with keyword arguments as parameters. Substitutes path variables in URI
         using path_vars. Raises ValueError if endpoint name is not found.
 
-        :rtype: JsonDict
+        Return type: JsonDict
         """
         path_vars = path_vars or {}
         # Load API endpoints once
@@ -325,35 +412,41 @@ class Tesla(requests.Session):
         if endpoint['AUTH'] and not self.authorized:
             self.fetch_token()
         # Substitute path variables in URI
-        uri = endpoint['URI']
         try:
-            uri = uri.format(**path_vars)
+            uri = endpoint['URI'].format(**path_vars)
         except KeyError as e:
             raise ValueError('%s requires path variable %s' % (name, e))
         # Perform request using given keyword arguments as parameters
+        if endpoint['TYPE'] == 'GET':
+            return self.request('GET', uri, params=kwargs)
         return self.request(endpoint['TYPE'], uri, data=kwargs)
 
     def vehicle_list(self):
-        """ Returns a list of :class: Vehicle <Vehicle> objects """
+        """ Returns a list of `Vehicle` objects """
         return [Vehicle(v, self) for v in self.api('VEHICLE_LIST')['response']]
-    
+
     def battery_list(self):
-        """ Returns a list of :class: Battery <Battery> objects """
+        """ Returns a list of `Battery` objects """
         return [Battery(p, self) for p in self.api('PRODUCT_LIST')['response']
                 if p.get('resource_type') == 'battery']
 
-    
-class HTMLForm(HTMLParser):
+    def solar_list(self):
+        """ Returns a list of `SolarPanel` objects """
+        return [SolarPanel(p, self) for p in self.api('PRODUCT_LIST')['response']
+                if p.get('resource_type') == 'solar']
+
+
+class HTMLForm(HTMLParser, dict):
     """ Parse input tags on HTML form """
 
-    def __init__(self):
-        self.data = {}
+    def __init__(self, html):
         HTMLParser.__init__(self)
+        self.feed(html)
 
     def handle_starttag(self, tag, attrs):
         """ Make dictionary of name and value attributes of input tags """
         if tag == 'input':
-            self.data[dict(attrs)['name']] = dict(attrs).get('value', '')
+            self[dict(attrs)['name']] = dict(attrs).get('value', '')
 
 
 class VehicleError(Exception):
@@ -373,10 +466,86 @@ class Vehicle(JsonDict):
     """ Vehicle class with dictionary access and API request support """
 
     codes = None  # Vehicle option codes class variable
+    COLS = ['speed', 'odometer', 'soc', 'elevation', 'est_heading', 'est_lat',
+            'est_lng', 'power', 'shift_state', 'range', 'est_range', 'heading']
 
     def __init__(self, vehicle, tesla):
         super(Vehicle, self).__init__(vehicle)
         self.tesla = tesla
+        self.callback = None
+
+    def _subscribe(self, wsapp):
+        """ Authenticate and select streaming telemetry columns """
+        msg = {'msg_type': 'data:subscribe_oauth',
+               'token': self.tesla.token['access_token'],
+               'value': ','.join(self.COLS), 'tag': str(self['vehicle_id'])}
+        wsapp.send(json.dumps(msg))
+
+    def _parse_msg(self, wsapp, message):
+        """ Parse messages """
+        msg = json.loads(message)
+        if msg['msg_type'] == 'control:hello':
+            logger.debug('connected')
+        elif msg['msg_type'] == 'data:update':
+            # Parse comma separated data record
+            data = dict(zip(['timestamp'] + self.COLS, msg['value'].split(',')))
+            for key, value in data.items():
+                try:
+                    data[key] = ast.literal_eval(value) if value else None
+                except ValueError:
+                    pass
+            logger.debug('Update %s', json.dumps(data))
+            if self.callback:
+                self.callback(data)
+            # Update polled data with streaming telemetry data
+            drive_state = self.setdefault('drive_state', JsonDict())
+            vehicle_state = self.setdefault('vehicle_state', JsonDict())
+            charge_state = self.setdefault('charge_state', JsonDict())
+            drive_state['timestamp'] = data['timestamp']
+            drive_state['speed'] = data['speed']
+            vehicle_state['odometer'] = data['odometer']
+            charge_state['battery_level'] = data['soc']
+            drive_state['heading'] = data['est_heading']
+            drive_state['latitude'] = data['est_lat']
+            drive_state['longitude'] = data['est_lng']
+            drive_state['power'] = data['power']
+            drive_state['shift_state'] = data['shift_state']
+            charge_state['ideal_battery_range'] = data['range']
+            charge_state['est_battery_range'] = data['est_range']
+            drive_state['heading'] = data['heading']
+        elif msg['msg_type'] == 'data:error':
+            logger.error(msg['value'])
+            wsapp.close()
+
+    @staticmethod
+    def _ws_error(wsapp, err):
+        """ Log exceptions """
+        logger.error(err)
+
+    def stream(self, callback=None, retry=0, indefinitely=False, **kwargs):
+        """ Let vehicle push on-change data, with 10 second idle timeout.
+
+        callback: (optional) Function with one argument, a dict of pushed data.
+        retry: (optional) Number of connection retries.
+        indefinitely: (optional) Retry indefinitely.
+        **kwargs: Optional arguments that `run_forever` takes.
+        """
+        self.callback = callback
+        websocket.enableTrace(logger.isEnabledFor(logging.DEBUG),
+                              handler=logging.NullHandler())
+        wsapp = websocket.WebSocketApp(STREAMING_BASE_URL + 'streaming/',
+                                       on_open=self._subscribe,
+                                       on_message=self._parse_msg,
+                                       on_error=self._ws_error)
+        kwargs.setdefault('ping_interval', 10)
+        while True:
+            wsapp.run_forever(**kwargs)
+            if indefinitely:
+                continue
+            if not retry:
+                break
+            logger.debug('%d retries left', retry)
+            retry -= 1
 
     def api(self, name, **kwargs):
         """ Endpoint request with vehicle_id path variable """
@@ -429,6 +598,12 @@ class Vehicle(JsonDict):
         """ Lists nearby Tesla-operated charging stations """
         return self.api('NEARBY_CHARGING_SITES')['response']
 
+    def get_service_scheduling_data(self):
+        """ Retrieves next service appointment for this vehicle """
+        response = self.api('SERVICE_SELF_SCHEDULING_ELIGIBILITY')['response']
+        return next((enabled for enabled in response['enabled_vins']
+                     if enabled['vin'] == self['vin']), {})
+
     def mobile_enabled(self):
         """ Checks if the Mobile Access setting is enabled in the car """
         # Construct URL and send request
@@ -509,17 +684,17 @@ class Vehicle(JsonDict):
             raise VehicleError(response['reason'])
         return response['result']
 
-    
-class BatteryError(Exception):
-    """ Battery exception class """
+
+class ProductError(Exception):
+    """ Product exception class """
     pass
 
 
-class Battery(JsonDict):
-    """ Battery class with dictionary access and API request support """
+class Product(JsonDict):
+    """ Base product class with dictionary access and API request support """
 
-    def __init__(self, battery, tesla):
-        super(Battery, self).__init__(battery)
+    def __init__(self, product, tesla):
+        super(Product, self).__init__(product)
         self.tesla = tesla
 
     def api(self, name, **kwargs):
@@ -527,18 +702,68 @@ class Battery(JsonDict):
         pathvars = {'battery_id': self['id'], 'site_id': self['energy_site_id']}
         return self.tesla.api(name, pathvars, **kwargs)
 
+    def get_calendar_history_data(
+            self, kind='savings', period='day', start_date=None,
+            end_date=time.strftime('%Y-%m-%dT%H:%M:%S.000Z'),
+            installation_timezone=None, timezone=None, tariff=None):
+        """ Retrieve live status of product
+        kind: A telemetry type of 'backup', 'energy', 'power',
+              'self_consumption', 'time_of_use_energy',
+              'time_of_use_self_consumption' and 'savings'
+        period: 'day', 'month', 'year', or 'lifetime'
+        end_date: The final day in the data requested in the json format
+                  '2021-02-28T07:59:59.999Z'
+        time_zone: Timezone in the json timezone format. eg. Europe/Brussels
+        start_date: The state date in the data requested in the json format
+                    '2021-02-27T07:59:59.999Z'
+        installation_timezone: Timezone of installation location for 'savings'
+        tariff: Unclear format use in 'savings' only
+        """
+        return self.api('CALENDAR_HISTORY_DATA', kind=kind, period=period,
+                        start_date=start_date, end_date=end_date,
+                        timezone=timezone,
+                        installation_timezone=installation_timezone,
+                        tariff=tariff)['response']
+
+    def get_history_data(
+            self, kind='savings', period='day', start_date=None,
+            end_date=time.strftime('%Y-%m-%dT%H:%M:%S.000Z'),
+            installation_timezone=None, timezone=None, tariff=None):
+        """ Retrieve live status of product
+        kind: A telemetry type of 'backup', 'energy', 'power',
+              'self_consumption', 'time_of_use_energy', and
+              'time_of_use_self_consumption'
+        period: 'day', 'month', 'year', or 'lifetime'
+        end_date: The final day in the data requested in the json format
+                  '2021-02-28T07:59:59.999Z'
+        time_zone: Timezone in the json timezone format. eg. Europe/Brussels
+        start_date: The state date in the data requested in the json format
+                    '2021-02-27T07:59:59.999Z'
+        installation_timezone: Timezone of installation location for 'savings'
+        tariff: Unclear format use in 'savings' only
+        """
+        return self.api('HISTORY_DATA', kind=kind, period=period,
+                        start_date=start_date, end_date=end_date,
+                        timezone=timezone,
+                        installation_timezone=installation_timezone,
+                        tariff=tariff)['response']
+
+    def command(self, name, **kwargs):
+        """ Wrapper method for product command response error handling """
+        response = self.api(name, **kwargs)['response']
+        if response['code'] == 201:
+            return response.get('message')
+        raise ProductError(response.get('message'))
+
+
+class Battery(Product):
+    """ Powerwall class """
+
     def get_battery_data(self):
         """ Retrieve detailed state and configuration of the battery """
         self.update(self.api('BATTERY_DATA')['response'])
         return self
 
-    def command(self, name, **kwargs):
-        """ Wrapper method for battery command response error handling """
-        response = self.api(name, **kwargs)['response']
-        if response['code'] == 201:
-            return response.get('message')
-        raise BatteryError(response.get('message'))
-
     def set_operation(self, mode):
         """ Set battery operation to self_consumption, backup or autonomous """
         return self.command('BATTERY_OPERATION_MODE', default_real_mode=mode)
@@ -548,3 +773,11 @@ class Battery(JsonDict):
         return self.command('BACKUP_RESERVE',
                             backup_reserve_percent=int(percent))
 
+
+class SolarPanel(Product):
+    """ Solar panel class """
+
+    def get_site_data(self):
+        """ Retrieve current site generation data """
+        self.update(self.api('SITE_DATA')['response'])
+        return self
